package mainForm;

import javafx.application.Platform;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.control.*;
import javafx.scene.control.Button;
import javafx.scene.control.TextField;
import javafx.stage.DirectoryChooser;

import java.awt.*;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ResourceBundle;
import java.util.function.IntConsumer;


public class Controller implements Initializable {
    private Model model = Model.getInstance();

    private Boolean isUpdating = false;

    @FXML
    private Spinner<Integer> spnBPM;
    @FXML
    private Spinner<Integer> spnBarCount;
    @FXML
    private ComboBox<String> cmbKey;
    @FXML
    private ComboBox<String> cmbMode;
    @FXML
    private TextField tfTotalLength;
    @FXML
    private ComboBox<String> cmbSaveAs;
    @FXML
    private Button btnSaveAs;
    @FXML
    private CheckBox chbCloseApp;
    @FXML
    private CheckBox chbOpenFolder;

    @FXML
    private void handleBtnSaveAsAction(){
        File file;
        Path path = Paths.get(model.getTargetFilePath());
        if (Files.isDirectory(path)) {
            file = new File(model.getTargetFilePath());
        } else if ((path.getParent() == null) ) {
            file = new File(model.getFilePaths().get(0));
        } else {
            file = new File(path.getParent().toString());
        }

        if (!file.exists())
            file = new File(model.getFilePaths().get(0));

        DirectoryChooser directoryChooser = new DirectoryChooser();
        directoryChooser.setTitle("Save As");
        directoryChooser.setInitialDirectory(file);
        file = directoryChooser.showDialog(btnSaveAs.getScene().getWindow());

        if (file != null) {
            model.setTargetFilePath(file.toString());
            cmbSaveAs.setValue(file.toString());
        }
    }

    @FXML
    private void handleBtnGenerateAction(){

        try {
            File file = new File(model.getTargetFilePath());
            if (!file.isDirectory()){
                file = file.getParentFile();
            }

            if (file == null){
                throw new FileNotFoundException("Invalid path!");
            }

            Generator generator = new Generator(model.getBpm(),
                    model.getBarCount(),
                    model.getKeys().indexOf(cmbKey.getValue()) -7,
                    (model.getModes().indexOf(cmbMode.getValue()) == 1));

            generator.generate(model.getTargetFilePath());

            if (model.getOpenFolder()){
                Desktop.getDesktop().open(file);
            }

            if (model.getCloseApp()){
                System.exit(0);
            }

            String str = Paths.get(model.getTargetFilePath()).getParent().toString();
            model.setTargetFilePath(model.targetDistinctFilePath(str));

        } catch (IOException e){
            showError(e.getMessage());
        } finally {
            updateForm();
        }
    }

    @FXML
    private void handleHplInfoAction(){
        Main.getMainHostServices().showDocument("https://github.com/MarekJurica/procedural_music_generation");
    }

    private void updateForm() {
        if (!isUpdating) {
            isUpdating = true;
            try {
                int a = 4 * model.getBarCount();
                int b = model.getBpm();

                double time = (double) a / b;
                int minutes = (int) Math.floor(time);
                int seconds = (int) Math.round((time - minutes) * 60);
                String text = Integer.toString(minutes) + ":";

                if (seconds < 9)
                    text = text + "0";

                text = text + Integer.toString(seconds);
                tfTotalLength.setText(text);

                cmbSaveAs.setValue(model.getTargetFilePath());
            }
            finally {
                isUpdating = false;
            }
        }
    }

    private void showError(String message){
        Alert alert = new Alert(Alert.AlertType.ERROR);
        alert.setTitle("Error");
        alert.setHeaderText(null);
        alert.setContentText(message);
        alert.showAndWait();
    }

    private void initSpinner(final Spinner spinner,
                             final int min,
                             final int max,
                             final int defaultValue,
                             final IntConsumer setter){
        if (min > max)
            throw new IllegalArgumentException("min is bigger than max");

        if (defaultValue < min)
            throw new IllegalArgumentException("defaultValue is smaller than min");

        if (defaultValue > max)
            throw new IllegalArgumentException("defaultValue is bigger than max");

        spinner.getEditor().textProperty().addListener((observable, oldValue, newValue) -> {
            if (!"".equals(newValue)) {
                try {
                    int old, val;
                    try {
                        old = Integer.parseInt(oldValue);
                    } catch (NumberFormatException nfe) {
                        old = defaultValue;
                    }

                    try {
                        val = Integer.parseInt(newValue);
                    } catch (NumberFormatException nfe) {
                        val = old;
                    }

                    if (val > max) {
                        val = max;
                    } else if (val < min) {
                        val = min;
                    }

                    if (!newValue.equals(Integer.toString(val)))
                        spinner.getEditor().setText(Integer.toString(val));

                    setter.accept(val);
                }
                finally {
                    updateForm();
                }
            }
        });
    }

    @FXML
    public void initialize(URL url, ResourceBundle rb) {
        try {
            spnBPM.setValueFactory(new SpinnerValueFactory.IntegerSpinnerValueFactory(Model.getMIN_BPM(),
                    Model.getMAX_BPM(),
                    model.getBpm()));

            initSpinner(spnBPM, Model.getMIN_BPM(), Model.getMAX_BPM(), model.getBpm(), (x) -> model.setBpm(x));

            ObservableList<Integer> barCountValues = FXCollections.observableArrayList();
            for (int i = Model.getMIN_BAR_COUNT(); i <= Model.getMAX_BAR_COUNT(); i = i + 8) {
                barCountValues.add(i);
            }
            SpinnerValueFactory<Integer> lVf = new SpinnerValueFactory.ListSpinnerValueFactory<>(barCountValues);
            lVf.setValue(model.getBarCount());
            spnBarCount.setValueFactory(lVf);

            initSpinner(spnBarCount,
                    Model.getMIN_BAR_COUNT(),
                    Model.getMAX_BAR_COUNT(),
                    model.getBarCount(),
                    (x) -> model.setBarCount(x));

            cmbKey.setItems(model.getKeys());
            cmbKey.getSelectionModel().select(model.getKeyIndex());

            cmbMode.setItems(model.getModes());
            cmbMode.getSelectionModel().select(model.getModeIndex());

            chbOpenFolder.selectedProperty().addListener((observable, oldValue, newValue) -> {
                try {
                    model.setOpenFolder(newValue);
                }
                finally {
                    updateForm();
                }
            });

            chbCloseApp.selectedProperty().addListener((observable, oldValue, newValue) -> {
                try {
                    model.setCloseApp(newValue);
                }
                finally {
                    updateForm();
                }
            });

            cmbSaveAs.setItems(model.getFilePaths());

            cmbSaveAs.getSelectionModel().selectedItemProperty().addListener((observable, oldValue, newValue) -> {
                Path newPath = Paths.get(newValue);
                String oldPathParentStr = "";
                if (oldValue != null) {
                    Path oldPathParent = Paths.get(oldValue).getParent();
                    if (oldPathParent != null)
                        oldPathParentStr = oldPathParent.toString();
                }

                if (Files.isDirectory(newPath) && (!newPath.toString().equals(oldPathParentStr))){
                    final String str = model.targetDistinctFilePath(newValue);
                    Platform.runLater(() -> cmbSaveAs.setValue(str)); // not so clean but quick
                    model.setTargetFilePath(str);
                }
            });

            cmbSaveAs.getEditor().textProperty().addListener((observable, oldValue, newValue) ->
                model.setTargetFilePath(newValue)
            );

            cmbSaveAs.getSelectionModel().selectFirst();
        }
        finally {
            updateForm();
        }
    }

}